/*
 *  SAP2
 *  version 2.0
 *  copyright (C) 2000-2001 Eric Botcazou
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*  version 1.0: programme fonctionnel sous MSDOS et Linux
 *          1.1: retour au format PC  la sortie du programme
 *          1.2: ajout d'un mode de fonctionnement batch
 *          2.0: support des disquettes 5"25 et de deux lecteurs PC
 */


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "libsap.h"
#include "to8.h"


#define SAP2_VERSION_STR "2.0"

#ifdef linux
   #define SAP2_PLATFORM_STR "Linux"
#else
   #define SAP2_PLATFORM_STR "MSDOS"
#endif

#define FILENAME_LENGTH 128


/* pour reproduire l'interface de Teo */
int (*to8_DirectReadSector)(int, int, int, int, unsigned char []);
int (*to8_DirectWriteSector)(int, int, int, int, const unsigned char []);
int (*to8_DirectFormatTrack)(int, int, const unsigned char []);

/* horrible hack pour supporter les accents... */
#ifdef linux
static char eacute[] = "";
static char egrave[] = "";
static char agrave[] = "";
static char ugrave[] = "";
#else
static char eacute[] = "";
static char egrave[] = "";
static char agrave[] = "";
static char ugrave[] = "";
#endif


/* pour l'affichage du rpertoire */
#define PAGE_HEIGHT 22


static int num_drives = 0;
static int drive_type[4];
static char *drive_type_name[7] = { "non disponible",
                                    "5\"25 - 360 ko",
                                    "5\"25 - 1.2 Mo",
                                    "3\"5 -  720 ko",
                                    "3\"5 - 1.44 Mo",
                                    "3\"5 - 2.88 Mo",
                                    "3\"5 - 2.88 Mo" };

#define IS_5_INCHES(drive) ((drive_type[drive]>0) && (drive_type[drive]<3))



/* term_puts:
 *  Affiche une chane de caractres en respectant la taille du terminal.
 */
static void term_puts(const char buffer[], int lines, int page_height)
{
   const char *start_cur, *end_cur;
   char trash[8];
   int i;

   start_cur = end_cur = buffer;

   while (lines>page_height) {
      for (i=0; i<page_height; i++) {
         while (*end_cur++ != '\n')
            ;
      }

      fwrite(start_cur, sizeof(char), end_cur-start_cur, stdout);

      fgets(trash, 7, stdin);
      printf("Appuyer sur <Return>");
      getchar();
      printf("\n");
      ungetc('\n', stdin);

      start_cur = end_cur;
      lines -= page_height;
   }

   puts(end_cur);
}



/* ViewDiskDir:
 *  Affiche le rpertoire d'une disquette TO.
 */
static int ViewDiskDir(int drive, int page_height)
{
   unsigned char data[512*SAP_NSECTS];
   char buffer[SAP_TRACKSIZE];
   int lines;

   if (to8_DirectReadSector(drive, 20, 1, SAP_NSECTS, data) != 0)
      return 1;

   /* on rduit la FAT  80 blocks pour les disquettes 5"25 */
   if (IS_5_INCHES(drive))
      memset(data + SAP_SECTSIZE + 1 + SAP_NTRACKS, 0xFE, SAP_NTRACKS);

   /* fonction bas-niveau de libSAP */
   lines = _ExtractDir(buffer, sizeof(buffer), drive, data);

   if ((page_height<0) || (lines<=page_height))
      puts(buffer);
   else
      term_puts(buffer, lines, page_height);

   return 0;
}



/* ViewArchiveDir:
 *  Affiche le rpertoire d'une archive SAP.
 */
static int ViewArchiveDir(const char sap_name[], int page_height)
{
   char buffer[SAP_TRACKSIZE];
   sapID sap_file;
   int lines;

   if ((sap_file=sap_OpenArchive(sap_name)) == SAP_ERROR)
      return 1;

   lines = sap_ListArchive(sap_file, buffer, sizeof(buffer));

   if (lines == 0)
      return 2;

   if ((page_height<0) || (lines<=page_height))
      puts(buffer);
   else
      term_puts(buffer, lines, page_height);

   sap_CloseArchive(sap_file);
   return 0;
}



/* CreateEmptyArchive:
 *  Cre une archive SAP vide (mais formate).
 */
static int CreateEmptyArchive(const char sap_name[], int format)
{
   sapID sap_file;

   if ((sap_file=sap_CreateArchive(sap_name)) == SAP_ERROR)
      return 1;

   sap_FormatArchive(sap_file, format);

   sap_CloseArchive(sap_file);

   return 0;
}



/* BuildSectorMap:
 *  Construit la carte des secteurs d'une piste
 *  en fonction du facteur d'entrelacement.
 */
static void BuildSectorMap(int *sector_map, int factor)
{
   int sect, loc=0;

   /* mise  zro de la table */
   memset(sector_map, 0, sizeof(int)*SAP_NSECTS);

   for (sect=1; sect<=SAP_NSECTS; sect++) {
      while (sector_map[loc] != 0)
         loc=(loc+1)%SAP_NSECTS;

      sector_map[loc]=sect;

      loc=(loc+factor)%SAP_NSECTS;
   }
}



/* FormatDisk:
 *  Formate une disquette 3"5 - 720 ko au format Thomson.
 */
static int FormatDisk(int drive, int factor, int verbose)
{
   int num_tracks;
   int track, sect, pos, i;
   int sector_map[SAP_NSECTS];
   unsigned char header_table[512];
   unsigned char buffer[512*SAP_NSECTS];

   /* dtermination du nombre de pistes */
   if (IS_5_INCHES(drive))
      num_tracks = SAP_NTRACKS/2;
   else
      num_tracks = SAP_NTRACKS;

   /* construction de la carte des secteurs pour chaque piste,
       partir du facteur d'entrelacement */
   BuildSectorMap(sector_map, factor);

   /* formatage des pistes */
   for (track=0; track<num_tracks; track++) {
      if (verbose) {
         printf("\r formatage piste %d", track);
         fflush(stdout);
      }

      /* construction de la table des headers */
      for (sect=1, pos=0; sect<=SAP_NSECTS; sect++, pos+=4) {
         header_table[pos]   = track;
         header_table[pos+1] = 0;
         header_table[pos+2] = sector_map[sect-1];
         header_table[pos+3] = 1;
      }

      if (to8_DirectFormatTrack(drive, track, header_table) != 0)
         return 1;
   }

   if (verbose) {
      printf("\n construction de la FAT (piste 20)");
      fflush(stdout);
   }

   for (i=0; i<SAP_SECTSIZE*SAP_NSECTS; i++)
      buffer[i] = 0xFF;

   /* pour MS-DOS (voir teo/src/dos/ddisk.c) */
   buffer[i]=0;

   if (to8_DirectWriteSector(drive, 20, 1, SAP_NSECTS, buffer) != 0)
      return 2;

   buffer[0] = 0;
   buffer[41] = 0xFE;
   buffer[42] = 0xFE;

   /* pour MS-DOS (voir teo/src/dos/ddisk.c) */
   buffer[SAP_SECTSIZE] = 0;

   if (to8_DirectWriteSector(drive, 20, 2, 1, buffer) != 0)
      return 2;

   if (verbose)
      printf("\n");

   return 0;
}



/* shrink_fat_to_80:
 *  Rduit la taille de la FAT  80 blocks (au lieu de 160).
 */
static void shrink_fat_to_80(unsigned char fat_data[])
{
   int i;

   for (i=0; i<SAP_NTRACKS; i++) {
      if (fat_data[1 + SAP_NTRACKS + i] != 0xFF)
         goto done;
   }

   memset(fat_data + 1 + SAP_NTRACKS, 0xFE, SAP_NTRACKS);

 done:
}



/* PackArchive:
 *  Transfre le contenu d'une disquette TO dans une archive SAP.
 */
static int PackArchive(const char sap_name[], int drive, int verbose)
{
   int num_tracks;
   int track, sect;
   sapID sap_file;
   sapsector_t sapsector;
   unsigned char buffer[512*SAP_NSECTS];

   if ((sap_file=sap_CreateArchive(sap_name)) == SAP_ERROR)
      return 1;

   /* dtermination du nombre de pistes */
   if (IS_5_INCHES(drive))
      num_tracks = SAP_NTRACKS/2;
   else
      num_tracks = SAP_NTRACKS;

   for (track=0; track<num_tracks; track++) {
      if (verbose) {
         printf("\r lecture piste %d", track);
         fflush(stdout);
      }

      if (to8_DirectReadSector(drive, track, 1, SAP_NSECTS, buffer) != 0) {
         /* erreur dans la piste, on essaie secteur par secteur */
         for (sect=1; sect<=SAP_NSECTS; sect++) {
            sapsector.format     = 0;
            sapsector.protection = 0;
            sapsector.track      = track;
            sapsector.sector     = sect;

            if (to8_DirectReadSector(drive, track, sect, 1, buffer) != 0) {
               /* erreur de lecture du secteur */
               if (verbose)
                  printf("\n   secteur %d illisible", sect);

               sapsector.format = 4;
               memset(sapsector.data, 0xF7, SAP_SECTSIZE);
            }
            else {
               memcpy(sapsector.data, buffer, SAP_SECTSIZE);
            }

            /* traitement spcial de la FAT pour les disquettes 5"25 */
            if ((track == 20) && (sect == 2) && IS_5_INCHES(drive))
               shrink_fat_to_80(sapsector.data);

            sap_FillArchive(sap_file, &sapsector);
         }

         if ((track < SAP_NTRACKS-1) && verbose)
            printf("\n");
      }
      else {
         for (sect=1; sect<=SAP_NSECTS; sect++) {
            sapsector.format     = 0;
            sapsector.protection = 0;
            sapsector.track      = track;
            sapsector.sector     = sect;

            memcpy(sapsector.data, buffer+(sect-1)*SAP_SECTSIZE, SAP_SECTSIZE);

            /* traitement spcial de la FAT pour les disquettes 5"25 */
            if ((track == 20) && (sect == 2) && IS_5_INCHES(drive))
               shrink_fat_to_80(sapsector.data);

            sap_FillArchive(sap_file, &sapsector);
         }
      }
   }

   if (verbose)
      printf("\n");

   sap_CloseArchive(sap_file);
   return 0;
}



/* UnpackArchive:
 *  Transfre le contenu d'une archive SAP vers une disquette TO.
 */
static int UnpackArchive(const char sap_name[], int drive, int verbose)
{
   int num_tracks;
   int track, sect;
   sapID sap_file;
   sapsector_t sapsector;
   unsigned char buffer[512*SAP_NSECTS];

   if ((sap_file=sap_OpenArchive(sap_name)) == SAP_ERROR)
      return 1;

   /* dtermination du nombre de pistes */
   if (IS_5_INCHES(drive))
      num_tracks = SAP_NTRACKS/2;
   else
      num_tracks = SAP_NTRACKS;

   for (track=0; track<num_tracks; track++) {
      if (verbose) {
         printf("\r %scriture piste %d", eacute, track);
         fflush(stdout);
      }

      for (sect=1; sect<=SAP_NSECTS; sect++) {
         sapsector.format     = 0;
         sapsector.protection = 0;
         sapsector.track      = track;
         sapsector.sector     = sect;

         if (sap_ReadSector(sap_file, track, sect, &sapsector) == SAP_ERROR) {
            sap_CloseArchive(sap_file);
            return 2;
         }
         else {
            memcpy(buffer+(sect-1)*SAP_SECTSIZE, sapsector.data, SAP_SECTSIZE);
         }
      }

      buffer[SAP_NSECTS*SAP_SECTSIZE] = 0;

      if (to8_DirectWriteSector(drive, track, 1, SAP_NSECTS, buffer) != 0) {
         sap_CloseArchive(sap_file);
         return 3;
      }
   }

   if (verbose)
      printf("\n");

   sap_CloseArchive(sap_file);
   return 0;
}



/* get_drive:
 *  Helper pour obtenir le numro de lecteur.
 */
static int get_drive(void)
{
   char trash[8];
   int drive;

   if (num_drives > 1) {
      do {
         printf(" num%sro du lecteur: ", eacute);
         if (!scanf("%d", &drive))
            fgets(trash, 7, stdin);
      }
      while ((drive<0) || (drive>3) || (drive_type[drive]==0));
   }
   else {
      for (drive=0; drive<4; drive++) {
         if (drive_type[drive] > 0)
            break;
      }
   }

   return drive;
}



/* interactive_main:
 *  Point d'entre du programme interactif.
 */
static void interactive_main(void)
{
   int c, ret;
   char trash[8];
   int drive;
   int factor = (SAP_NSECTS-1)/2;
   char sap_name[FILENAME_LENGTH];

   while (1) {
      printf("SAP2, syst%sme d'archivage pour disquettes 3\"5 et 5\"25 Thomson\n", egrave);
      printf("version "SAP2_VERSION_STR" ("SAP2_PLATFORM_STR") copyright (C) 2000-2001 Eric Botcazou\n");
      printf(" bas%s sur SAP copyright (C) Alexandre Pukall Avril 1998\n\n", eacute);

      if (drive_type[0] > 0) {
         printf("Lecteur A: (%s) --> lecteur 0", drive_type_name[drive_type[0]]);

         if (drive_type[1] > 0)
            printf(" + lecteur 1\n");
         else
            printf("\n");
      }
      else {
         printf("Lecteur A: (%s)\n", drive_type_name[0]);
      }

      if (drive_type[2] > 0) {
         printf("Lecteur B: (%s) --> lecteur 2", drive_type_name[drive_type[2]]);

         if (drive_type[3] > 0)
            printf(" + lecteur 3\n\n");
         else
            printf("\n\n");
      }
      else {
         printf("Lecteur B: (%s)\n\n", drive_type_name[0]);
      }

      printf("Transfert TO-->PC:\n");
      printf(" 1. Visualiser le contenu d'une disquette Thomson\n");
      printf(" 2. Cr%ser une archive SAP vide\n", eacute);
      printf(" 3. Archiver une disquette Thomson vers une archive SAP\n");
      printf("Transfert PC-->TO:\n");
      printf(" 4. Visualiser le contenu d'une archive SAP\n");
      printf(" 5. Formater une disquette 3\"5 (720 ko) ou 5\"25 (360 ko) au format Thomson\n");
      printf(" 6. D%ssarchiver une archive SAP vers une disquette Thomson\n", eacute);
      printf("Autres commandes:\n");
      printf(" 7. Quitter\n\n");

      do {
         printf("Votre choix: ");
         if (!scanf("%d", &c))
            fgets(trash, 7, stdin);
      }
      while ((c<1) || (c>7));

      switch (c) {

         case 1:
            printf("Visualisation du contenu d'une disquette:\n");
            drive = get_drive();
            printf("\n");

            ret = ViewDiskDir(drive, PAGE_HEIGHT);
            if (ret == 1)
               printf("*** Erreur: r%spertoire illisible ***\n", eacute);

            break;

         case 2:
            printf("Cr%sation d'une archive vide:\n", eacute);

            printf(" nom de l'archive %s cr%ser (sans extension): ", agrave, eacute);
            scanf("%s", sap_name);
            strcat(sap_name, ".sap");

            do {
               printf(" format (1->%d pistes, 2->%d pistes): ", SAP_NTRACKS, SAP_NTRACKS/2);
               if (!scanf("%d", &factor))
                  fgets(trash, 7, stdin);
            }
            while ((factor<1) || (factor>2));

            ret = CreateEmptyArchive(sap_name, factor == 1 ? SAP_TRK80 : SAP_TRK40);
            if (ret == 1)
               printf("*** Erreur: impossible de cr%ser le fichier %s ***\n", eacute, sap_name);

            break;

         case 3:
            printf("Archivage d'une disquette:\n");

            printf(" nom de l'archive %s cr%ser (sans extension): ", agrave, eacute);
            scanf("%s", sap_name);
            strcat(sap_name, ".sap");

            ret = PackArchive(sap_name, get_drive(), 1);
            if (ret == 1)
               printf("*** Erreur: impossible de cr%ser le fichier %s ***\n", eacute, sap_name);

            break;

         case 4:
            printf("Visualisation du contenu d'une archive:\n");

            printf(" nom de l'archive (sans extension): ");
            scanf("%s", sap_name);
            strcat(sap_name, ".sap");

            printf("\n");

            ret = ViewArchiveDir(sap_name, PAGE_HEIGHT);
            if (ret == 1)
               printf("*** Erreur: impossible d'ouvrir le fichier %s ***\n", sap_name);
            else if (ret == 2)
               printf("*** Erreur: archive SAP corrompue ***\n");

            break;

         case 5:
            printf("Formatage d'une disquette au format Thomson:\n");
            printf(" Si la disquette est une 3\"5 - 1.44 Mo, occulter l'encoche\n");
            printf(" de droite (recto et verso) avec un morceau de scotch.\n");

            drive = get_drive();

            do {
               printf(" facteur d'entrelacement ([1..%d], %d recommand%s): ", SAP_NSECTS-1, (SAP_NSECTS-1)/2, eacute);
               if (!scanf("%d", &factor))
                  fgets(trash, 7, stdin);
            }
            while ((factor<1) || (factor>(SAP_NSECTS-1)));

            ret = FormatDisk(drive, factor, 1);
            if (ret == 1)
               printf("\n*** Erreur: impossible de formater la disquette ***\n");
            else if (ret == 2)
               printf("*** Erreur: impossible d'%scrire sur la disquette ***\n", eacute);
            break;

         case 6:
            printf("D%ssarchivage vers une disquette:\n", eacute);

            drive = get_drive();

            printf(" nom de l'archive (sans extension): ");
            scanf("%s", sap_name);
            strcat(sap_name, ".sap");

            ret = UnpackArchive(sap_name, drive, 1);
            if (ret == 1)
               printf("*** Erreur: impossible d'ouvrir le fichier %s ***\n", sap_name);
            else if (ret == 2)
               printf("*** Erreur: archive SAP corrompue ***\n");
            else if (ret == 3)
               printf("\n*** Erreur: impossible d'%scrire sur la disquette ***\n", eacute);
            break;

         case 7:
            return;
      }

      fgets(trash, 7, stdin);
      printf("Appuyer sur <Return>");
      fflush(stdout);
      getchar();
      printf("\n");
   }
}


#define COMMAND_MAX  8

static char *short_command[] = {"-h", "-v", "-d", "-c", "-p", "-t", "-f", "-u"};
static char *long_command[] = {"--help", "--version", "--dir", "--create", "--pack", "--list", "--format", "--unpack"};



/* usage:
 *  Affiche les commandes d'utilisation du programme et quitte.
 */
static void usage(void)
{
   fprintf(stderr, "Usage: sap2  -h --help | -v --version | -d --dir | -c --create | -p --pack |\n");
   fprintf(stderr, " -t --list | -f --format | -u --unpack  [archive.sap] [lecteur] [entrelacement]\n");
   exit(EXIT_FAILURE);
}



/* main:
 *  Point d'entre du programme.
 */
int main(int argc, char *argv[])
{
   int ret = 0;
   int i;

   if (argc < 2) { /* no argument? */
      if ((num_drives=InitDirectDisk(drive_type, 1)) > 0) {
         interactive_main();
         ExitDirectDisk();
      }
      else {
         ret = 1;
      }
   }
   else {
      if (argv[1][0] == '-') {

         switch (argv[1][1]) {

            case '-':  /* long commands */
               for (i=0; i<COMMAND_MAX; i++) {
                  if (strcmp(argv[1], long_command[i]) == 0) {
                     argv[1] = short_command[i];
                     return main(argc, argv);
                  }
               }

               usage();

            case 'h':  /* help */
               printf("SAP2 est un syst%sme d'archivage pour disquettes 3\"5 et 5\"25 Thomson. Il permet\n", egrave);
               printf("de cr%ser des images de disquettes sous forme d'archive SAP.\n\n", eacute);
               printf("Usage:\n");
               printf("      %s commande [archive.sap] [nb pistes] [lecteur] [entrelacement]\n", argv[0]);
               printf("o%s la commande est prise parmi les suivantes:\n", ugrave);
               printf("  -h, --help          affiche cette aide\n");
               printf("  -v, --version       affiche la version du programme\n");
               printf("  -d, --dir           affiche le contenu d'une disquette Thomson\n");
               printf("  -c, --create        cr%se une archive SAP vide\n", eacute);
               printf("  -p, --pack          archive une disquette Thomson vers une archive SAP\n");
               printf("  -t, --list          affiche la liste des fichiers de l'archive SAP\n");
               printf("  -f, --format        formate une disquette au format Thomson\n");
               printf("  -u, --unpack        d%ssarchive une archive SAP vers une disquette Thomson\n", eacute);
               break;

            case 'v':  /* version */
               printf("SAP2 version "SAP2_VERSION_STR" pour "SAP2_PLATFORM_STR", copyright (C) 2000-2001 Eric Botcazou.\n");
               break;

            case 'd':  /* dir */
               if (argc < 3)
                  usage();

               if (InitDirectDisk(drive_type, 1) > 0) {
                  ret = ViewDiskDir(atoi(argv[2]), -1);
                  ExitDirectDisk();
               }
               else {
                  ret = 1;
               }
               break;

            case 'c':  /* create */
               if (argc < 3)
                  usage();

               if (argc < 4) {
                  ret = CreateEmptyArchive(argv[2], SAP_TRK80);
               }
               else {
                  if (atoi(argv[3]) == 80)
                     ret = CreateEmptyArchive(argv[2], SAP_TRK80);
                  else if (atoi(argv[3]) == 40)
                     ret = CreateEmptyArchive(argv[2], SAP_TRK40);
                  else
                     ret = 1;
               }
               break;

            case 'p':  /* pack */
               if (argc < 4)
                  usage();

               if (InitDirectDisk(drive_type, 1) > 0) {
                  ret = PackArchive(argv[2], atoi(argv[3]), 0);
                  ExitDirectDisk();
               }
               else {
                  ret = 1;
               }
               break;

            case 't':  /* list */
               if (argc < 3)
                  usage();

               ret = ViewArchiveDir(argv[2], -1);
               break;

            case 'f':  /* format */
               if (argc < 3)
                  usage();

               if (InitDirectDisk(drive_type, 1) > 0) {
                  if (argc == 3)
                     ret = FormatDisk(atoi(argv[2]), (SAP_NSECTS-1)/2, 0);
                  else
                     ret = FormatDisk(atoi(argv[2]), atoi(argv[3]), 0);

                  ExitDirectDisk();
               }
               else {
                  ret = 1;
               }
               break;

            case 'u':  /* unpack */
               if (argc < 4)
                  usage();

               if (InitDirectDisk(drive_type, 1) > 0) {
                  ret = UnpackArchive(argv[2], atoi(argv[3]), 0);
                  ExitDirectDisk();
               }
               else {
                  ret = 1;
               }
               break;

            default:
               usage();
         }  /* end of switch */
      }
      else {
         usage();
      }
   }

   return ret;
}

