/*
 *  SAPfs
 *  version 0.9.1
 *  copyright (C) 2001 Eric Botcazou
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*  version 1.0:
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include "libsap.h"


#define SAPFS_VERSION_STR "0.9.1"

#ifdef linux
   #define SAPFS_PLATFORM_STR "Linux"
#else
   #define SAPFS_PLATFORM_STR "MSDOS"
#endif

#define FILENAME_LENGTH 512


/* ugly hack to support French accents */
#ifdef linux
static char eacute[] = "";
static char egrave[] = "";
static char agrave[] = "";
static char ugrave[] = "";
#else
static char eacute[] = "";
static char egrave[] = "";
static char agrave[] = "";
static char ugrave[] = "";
#endif



/* PrintErrorMessage:
 *  Prints the error message corresponding to the specified error.
 */
static void PrintErrorMessage(int errno, const char str[])
{
   switch (errno) {

      case SAP_EBADF:
         fprintf(stderr, "Erreur: le fichier %s n'est pas une archive SAP valide.\n", str);
         break;

      case SAP_EFBIG:
         fprintf(stderr, "Erreur: le fichier %s est de taille trop importante.\n", str);
         break;

      case SAP_ENFILE:
         fprintf(stderr, "Erreur: le fichier %s est vide.\n", str);
         break;

      case SAP_ENOENT:
         fprintf(stderr, "Erreur: le fichier %s est introuvable.\n", str);
         break;

      case SAP_ENOSPC:
         fprintf(stderr, "Erreur: le r%spertoire de l'archive est plein.\n", eacute);
         break;

      case SAP_EPERM:
         fprintf(stderr, "Erreur: impossible de cr%ser le fichier %s.\n", eacute, str);
         break;
   }
}



/* put_separator:
 *  Helper function to add a missing directory separator.
 */
static void put_separator(char filename[])
{
   int last;

   last = strlen(filename) - 1;

   if (filename[last] != '/')
      filename[last + 1] = '/';
}



/* CreateEmptyArchive:
 *  Creates a new archive using the specified name.
 */
static int CreateEmptyArchive(const char sap_name[], int format)
{
   sapID sap_file;

   if ((sap_file=sap_CreateArchive(sap_name)) == SAP_ERROR) {
      PrintErrorMessage(sap_errno, sap_name);
      return 1;
   }

   sap_FormatArchive(sap_file, format);
   sap_CloseArchive(sap_file);

   return 0;
}



/* FormatArchive:
 *  Formats the specified already existing archive.
 */
static int FormatArchive(const char sap_name[], int format)
{
   sapID sap_file;

   if ((sap_file=sap_OpenArchive(sap_name)) == SAP_ERROR) {
      PrintErrorMessage(sap_errno, sap_name);
      return 1;
   }

   sap_FormatArchive(sap_file, format);
   sap_CloseArchive(sap_file);

   return 0;
}



/* VerifyArchive:
 *  Verifies the specified archive.
 */
static int VerifyArchive(const char sap_name[])
{
   sapsector_t sapsector;
   sapID sap_file;
   int track, sect;
   int flag;

   if ((sap_file=sap_OpenArchive(sap_name)) == SAP_ERROR) {
      PrintErrorMessage(sap_errno, sap_name);
      return 1;
   }

   for (track=0; track<SAP_NTRACKS; track++) {
      for (sect=1; sect<=SAP_NSECTS; sect++) {
         flag = sap_ReadSector(sap_file, track, sect, &sapsector);

         if (flag != SAP_OK) {
            printf("track %d sector %02d: ", track, sect);

            if (flag & SAP_NO_STD_FMT)
               printf("<format=%d> ", sapsector.format);

            if (flag & SAP_PROTECTED)
               printf("<protection=%d> ", sapsector.protection);

            if (flag & SAP_BAD_SECTOR)
               printf("<track=%d sector=%d> ", sapsector.track, sapsector.sector);

            if (flag & SAP_CRC_ERROR)
               printf("<CRC error>");

            printf("\n");
         }
      } /* end of sect loop */
   } /* end of track loop */

   sap_CloseArchive(sap_file);

   return 0;
}



/* DumpArchive:
 *  Dumps one or more sectors from the specified archive.
 */
static int DumpArchive(const char sap_name[], int track, int sect)
{
   sapsector_t sapsector;
   unsigned char c;
   sapID sap_file;
   int i,j,s,t;

   if ((sap_file=sap_OpenArchive(sap_name)) == SAP_ERROR) {
      PrintErrorMessage(sap_errno, sap_name);
      return 1;
   }

   /* check track number */
   if (track<0) {
      t = 0;
      track = SAP_NTRACKS-1;
   }
   else {
      if (track>=SAP_NTRACKS) {
         fprintf(stderr, "Erreur: num%sro de piste invalide.\n", eacute);
         sap_CloseArchive(sap_file);
         return 1;
      }

      t = track;
   }

   /* check sector number */
   if (sect<0) {
      s = 1;
      sect = SAP_NSECTS;
   }
   else {
      if ((sect<1) || (sect>SAP_NSECTS)) {
         fprintf(stderr, "Erreur: num%sro de secteur invalide.\n", eacute);
         sap_CloseArchive(sap_file);
         return 1;
      }

      s = sect;
   }

   while (t <= track) {
      while (s <= sect) {
         sap_ReadSector(sap_file, t++, s++, &sapsector);

         printf("track id: %02d\n", sapsector.track);
         printf("sector id: %02d\n", sapsector.sector);
         printf("format: %d\n", sapsector.format);
         printf("protection: %d\n", sapsector.protection);
         printf("data:\n");

         for (i=0; i<16; i++) {
            for (j=0; j<16; j++) {
               c = sapsector.data[i*16+j];
               printf("%02X ", c);
            }

            for (j=0; j<16; j++) {
               c = sapsector.data[i*16+j];

               if ((c>=32) && (c<=125))
                  putchar(c);
               else
                  putchar('.');
            }

            printf("\n");
         }
      }  /* end of while (s <= sect) */
   }  /* end of while (t <= track ) */

   sap_CloseArchive(sap_file);

   return 0;
}



/* ListArchive:
 *  Lists the files contained in the specified archive.
 */
static int ListArchive(const char sap_name[])
{
   char buffer[4096];
   sapID sap_file;

   if ((sap_file=sap_OpenArchive(sap_name)) == SAP_ERROR) {
      PrintErrorMessage(sap_errno, sap_name);
      return 1;
   }

   sap_ListArchive(sap_file, buffer, sizeof(buffer));
   puts(buffer);
   sap_CloseArchive(sap_file);

   return 0;
}



/* ExtractFile:
 *  Extracts one or more files from the specified archive.
 */
static int ExtractFile(const char sap_name[], char *filename[], int nfiles)
{
   int i, len, ret = 0;
   sapID sap_file;

   if ((sap_file=sap_OpenArchive(sap_name)) == SAP_ERROR) {
      PrintErrorMessage(sap_errno, sap_name);
      return 1;
   }

   for (i=0; i<nfiles; i++) {
      len = sap_ExtractFile(sap_file, filename[i]);

      if (len == 0) {
         PrintErrorMessage(sap_errno, filename[i]);
         ret = 1;
         break;
      }
   }

   sap_CloseArchive(sap_file);

   return ret;
}



/* AddFile:
 *  Adds one or more files to the specified archive.
 */
static int AddFile(const char sap_name[], char *filename[], int nfiles)
{
   char pathname[FILENAME_LENGTH];
   int i, len = 0, ret = 0;
   struct stat s;
   struct dirent *entry;
   sapID sap_file;
   DIR *dir;

   if ((sap_file=sap_OpenArchive(sap_name)) == SAP_ERROR) {
      PrintErrorMessage(sap_errno, sap_name);
      return 1;
   }

   for (i=0; i<nfiles; i++) {
      /* get file statistics */
      if (stat(filename[i], &s) != 0) {
         fprintf(stderr, "Erreur: le fichier %s est introuvable.\n", filename[i]);
         ret = 1;
         break;
      }

      if (S_ISDIR(s.st_mode)) {  /* directory? */
         if ((dir=opendir(filename[i])) == NULL) {
            fprintf(stderr, "Erreur: le r%spertoire %s est inaccessible en lecture.\n", eacute, filename[i]);
            ret = 1;
            break;
         }

         /* add every entry in turn */
         while ((entry = readdir(dir)) != NULL) {
            if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, "..")) {
               strncpy(pathname, filename[i], FILENAME_LENGTH - 1);
               put_separator(pathname);
               strncat(pathname, entry->d_name, FILENAME_LENGTH - strlen(pathname) - 1);
               len = sap_AddFile(sap_file, pathname);

               if (len == 0) {
                  PrintErrorMessage(sap_errno, pathname);
                  ret = 1;
                  break;
               }
            }
         }

         closedir(dir);
      }
      else {
         len = sap_AddFile(sap_file, filename[i]);
      }

      if (len == 0) {
         if (ret == 0) {
            PrintErrorMessage(sap_errno, filename[i]);
            ret = 1;
         }
         break;
      }
   }

   sap_CloseArchive(sap_file);

   return ret;
}



/* DeleteFile:
 *  Deletes one or more files from the specified archive.
 */
static int DeleteFile(const char sap_name[], char *filename[], int nfiles)
{
   int i, len, ret = 0;
   sapID sap_file;

   if ((sap_file=sap_OpenArchive(sap_name)) == SAP_ERROR) {
      PrintErrorMessage(sap_errno, sap_name);
      return 1;
   }

   for (i=0; i<nfiles; i++) {
      len = sap_DeleteFile(sap_file, filename[i]);

      if (len == 0) {
         PrintErrorMessage(sap_errno, filename[i]);
         ret = 1;
         break;
      }
   }

   sap_CloseArchive(sap_file);

   return ret;
}


#define COMMAND_MAX  11

static char *short_command[] = { "-h", "-v", "-c", "-f", "-w", "-u", "-t", "-x", "-y", "-a", "-d" };
static char *long_command[] = { "--help", "--version", "--create", "--format", "--verify", "--dump",
                                "--list", "--extract", "--extract-all", "--add", "--delete" };


/* usage:
 *  Displays the commands and quits.
 */
static void usage(void)
{
   fprintf(stderr, "Usage:  sapfs -h --help | -v --version | -c --create |  -f --format |\n");
   fprintf(stderr, "              -w --verify | -d --dump | -t --list | -x --extract[-all] |\n");
   fprintf(stderr, "              -a --add | -d --delete\n");
   fprintf(stderr, "              archive.sap [fichier...] [nb pistes] [piste] [secteur]\n");
   exit(EXIT_FAILURE);
}



/* main:
 *  Entry point for the program.
 */
int main(int argc, char *argv[])
{
   int i, ret = 0;
   char *star = "*";

   if (argc < 2)  /* no argument? */
      usage();

   if (argv[1][0] == '-') {

      switch (argv[1][1]) {

         case '-':  /* long commands */
            for (i=0; i<COMMAND_MAX; i++) {
               if (strcmp(argv[1], long_command[i]) == 0) {
                  argv[1] = short_command[i];
                  return main(argc, argv);
               }
            }

            usage();

         case 'h':  /* help */
            printf("SAPfs est un outil de manipulation des archives SAP qui permet de r%saliser\n", eacute);
            printf("sur ces archives les op%srations naturelles d'un syst%sme de fichiers.\n\n", eacute, egrave);
            printf("Usage:\n");
            printf("    %s commande archive.sap [fichier...] [nb pistes] [piste] [secteur]\n", argv[0]);
            printf("o%s la commande est prise parmi les suivantes:\n", ugrave);
            printf("  -h, --help          affiche cette aide\n");
            printf("  -v, --version       affiche la version du programme\n");
            printf("  -c, --create        cr%se une archive SAP vide\n", eacute);
            printf("  -f, --format        formate une archive SAP\n");
            printf("  -w, --verify        effectue une v%srification de l'archive SAP\n", eacute);
            printf("  -u, --dump          affiche le contenu d'un ou plusieurs secteurs\n");
            printf("  -t, --list          affiche la liste des fichiers de l'archive SAP\n");
            printf("  -x, --extract       extrait un ou plusieurs fichiers de l'achive SAP\n");
            printf("      --extract-all   extrait tous les fichiers de l'archive SAP\n");
            printf("  -a, --add           ajoute un ou plusieurs fichiers %s l'archive SAP\n", agrave);
            printf("  -d, --delete        d%struit un ou plusieurs fichiers de l'archive SAP\n", eacute);
            break;

         case 'v':  /* version */
            printf("SAPfs version "SAPFS_VERSION_STR" pour "SAPFS_PLATFORM_STR", copyright (C) 2001 Eric Botcazou.\n");
            break;

         case 'c':  /* create */
            if (argc < 3)
               usage();

            if (argc < 4) {
               ret = CreateEmptyArchive(argv[2], SAP_TRK80);
            }
            else {
               if (atoi(argv[3]) == 80)
                  ret = CreateEmptyArchive(argv[2], SAP_TRK80);
               else if (atoi(argv[3]) == 40)
                  ret = CreateEmptyArchive(argv[2], SAP_TRK40);
               else
                  ret = 1;
            }
            break;

         case 'f':  /* format */
            if (argc < 3)
               usage();

            if (argc < 4) {
               ret = FormatArchive(argv[2], SAP_TRK80);
            }
            else {
               if (atoi(argv[3]) == 80)
                  ret = FormatArchive(argv[2], SAP_TRK80);
               else if (atoi(argv[3]) == 40)
                  ret = FormatArchive(argv[2], SAP_TRK40);
               else
                  ret = 1;
            }
            break;

         case 'w':  /* verify */
            if (argc < 3)
               usage();

            ret = VerifyArchive(argv[2]);
            break;


         case 'u':  /* dump */
            if (argc < 3)
               usage();

            if (argc < 4)
               ret = DumpArchive(argv[2], -1, -1);
            else if (argc < 5)
               ret = DumpArchive(argv[2], atoi(argv[3]), -1);
            else
               ret = DumpArchive(argv[2], atoi(argv[3]), atoi(argv[4]));
            break;

         case 't':  /* list */
            if (argc < 3)
               usage();

            ret = ListArchive(argv[2]);
            break;

         case 'x':  /* extract */
            if (argc < 4)
               usage();

            ret = ExtractFile(argv[2], argv+3, argc-3);
            break;

         case 'y':  /* extract all */
            if (argc < 3)
               usage();

            ret = ExtractFile(argv[2], &star, 1);
            break;

         case 'a':  /* add */
            if (argc < 4)
               usage();

            ret = AddFile(argv[2], argv+3, argc-3);
            break;

         case 'd':  /* delete */
            if (argc < 4)
               usage();

            ret = DeleteFile(argv[2], argv+3, argc-3);
            break;

         default:
            usage();
      }  /* end of switch */
   }

   return ret;
}

